////
// 🦠 Corona-Warn-App
//

import UIKit

@IBDesignable
class GradientView: UIView {

	// MARK: - Init

	init(
		type: GradientType = .solidGrey,
		withStars starsAreVisible: Bool = false,
		frame: CGRect = .zero
	) {
		super.init(frame: frame)

		setupView()

		self.type = type
		self.starsAreVisible = starsAreVisible

		updateLayer()
	}

	override init(frame: CGRect) {
		super.init(frame: frame)
		setupView()
		updateLayer()
	}

	required init?(coder: NSCoder) {
		super.init(coder: coder)
		setupView()
		updateLayer()
	}

	// MARK: - Overrides

	override class var layerClass: AnyClass {
		return CAGradientLayer.self
	}

	override func traitCollectionDidChange(_ previousTraitCollection: UITraitCollection?) {
		super.traitCollectionDidChange(previousTraitCollection)
		updateLayer()
	}

	// MARK: - Internal

	enum GradientType: Equatable {
		case blueRedTilted
		case blueOnly
		case solidGrey
		case lightBlue
		case mediumBlue
		case darkBlue
		case solidLightGreen
		case solidDarkGreen
		case green
		case whiteWithGreyBorder
		case whiteToLightBlue
		case lightBlueToWhite

		
		var starsColor: UIColor? {
			switch self {
			case .solidGrey:
				return UIColor(red: 87.0 / 255.0, green: 103.0 / 255.0, blue: 120.0 / 255.0, alpha: 1.0)
			case .lightBlue:
				return UIColor(red: 1.0 / 255.0, green: 145.0 / 255.0, blue: 198.0 / 255.0, alpha: 1.0)
			case .mediumBlue:
				return UIColor(red: 7.0 / 255.0, green: 106.0 / 255.0, blue: 159.0 / 255.0, alpha: 1.0)
			case .darkBlue:
				return UIColor(red: 2.0 / 255.0, green: 90.0 / 255.0, blue: 143.0 / 255.0, alpha: 1.0)
			case .green:
				return UIColor(red: 47.0 / 255.0, green: 137.0 / 255.0, blue: 77.0 / 255.0, alpha: 1.0)
			default:
				return nil
			}
		}
	}

	var type: GradientType = .blueRedTilted {
		didSet {
			updateLayer()
		}
	}

	var starsAreVisible: Bool = false {
		didSet {
			updateLayer()
		}
	}

	// MARK: - Private

	private let imageView: UIImageView = UIImageView()
	
	private var isDarkMode: Bool {
		if #available(iOS 13, *) {
			return UITraitCollection.current.userInterfaceStyle == .dark
		} else {
			return false
		}
	}

	private func setupView() {
		imageView.translatesAutoresizingMaskIntoConstraints = false
		addSubview(imageView)
		NSLayoutConstraint.activate(
			[
				imageView.topAnchor.constraint(equalTo: topAnchor, constant: 10.0),
				imageView.trailingAnchor.constraint(equalTo: trailingAnchor, constant: -33.0)
			]
		)
	}

	private func updateStars() {
		// update stars view
		if let starsColor = type.starsColor, starsAreVisible {
			imageView.tintColor = starsColor
			imageView.tintAdjustmentMode = .normal
			imageView.image = UIImage(imageLiteralResourceName: "EUStarsGroup")
		} else {
			imageView.image = nil
		}
	}

	// swiftlint:disable cyclomatic_complexity
	// swiftlint:disable function_body_length
	private func updateGradient() {
		guard let gradientLayer = self.layer as? CAGradientLayer else {
			Log.debug("Failed to create view with matching layer class", log: .default)
			return
		}
		// update gradient layer
		switch type {
		case .blueRedTilted:
			// magic numbers to create the gradient colors in the right place
			gradientLayer.colors = [
				UIColor(red: 0.235, green: 0.547, blue: 0.733, alpha: 1).cgColor,
				UIColor(red: 0.424, green: 0.392, blue: 0.55, alpha: 1).cgColor,
				UIColor(red: 0.663, green: 0.246, blue: 0.271, alpha: 1).cgColor
			]
			gradientLayer.locations = [0.12, 0.48, 0.81]
			gradientLayer.startPoint = CGPoint(x: 0.0, y: 0.5)
			gradientLayer.endPoint = CGPoint(x: 1.0, y: 1.0)

		case .blueOnly:
			gradientLayer.colors = [
				UIColor(red: 43 / 255, green: 84 / 255, blue: 142 / 255, alpha: 1).cgColor,
				UIColor(red: 29 / 255, green: 78 / 255, blue: 125 / 255, alpha: 1).cgColor,
				UIColor(red: 16 / 255, green: 62 / 255, blue: 110 / 255, alpha: 1).cgColor
			]
			gradientLayer.locations = [0.0, 0.5, 1.0]
			gradientLayer.startPoint = CGPoint(x: 0.0, y: 0.0)
			gradientLayer.endPoint = CGPoint(x: 1.0, y: 1.0)

		case .solidGrey:
			gradientLayer.colors = [
				UIColor(red: 0.38, green: 0.435, blue: 0.494, alpha: 1).cgColor,
				UIColor(red: 0.38, green: 0.435, blue: 0.494, alpha: 1).cgColor
			]
			gradientLayer.locations = [0.0, 1.0]
			gradientLayer.startPoint = CGPoint(x: 0.0, y: 0.0)
			gradientLayer.endPoint = CGPoint(x: 1.0, y: 1.0)

		case .lightBlue:
			gradientLayer.colors = [
				UIColor(red: 0, green: 0.575, blue: 0.783, alpha: 1).cgColor,
				UIColor(red: 0, green: 0.499, blue: 0.679, alpha: 1).cgColor
			]
			gradientLayer.locations = [0.0, 1.0]
			gradientLayer.startPoint = CGPoint(x: 0.25, y: 0.5)
			gradientLayer.endPoint = CGPoint(x: 0.75, y: 0.5)

		case .mediumBlue:
			gradientLayer.colors = [
				UIColor(red: 25 / 255, green: 108 / 255, blue: 163 / 255, alpha: 1).cgColor,
				UIColor(red: 35 / 255, green: 118 / 255, blue: 169 / 255, alpha: 1).cgColor
			]
			gradientLayer.locations = [0.0, 1.0]
			gradientLayer.startPoint = CGPoint(x: 0.0, y: 0.0)
			gradientLayer.endPoint = CGPoint(x: 1.0, y: 0.5)

		case .darkBlue:
			gradientLayer.colors = [
				UIColor(red: 0 / 255, green: 93 / 255, blue: 147 / 255, alpha: 1).cgColor,
				UIColor(red: 4 / 255, green: 96 / 255, blue: 151 / 255, alpha: 1).cgColor
			]
			gradientLayer.locations = [0.25, 0.75]
			gradientLayer.startPoint = CGPoint(x: 0.0, y: 0.0)
			gradientLayer.endPoint = CGPoint(x: 1.0, y: 1.0)

		case .solidLightGreen:
			gradientLayer.colors = [
				UIColor(red: 46 / 255, green: 133 / 255, blue: 75 / 255, alpha: 1).cgColor,
				UIColor(red: 46 / 255, green: 133 / 255, blue: 75 / 255, alpha: 1).cgColor
			]
			gradientLayer.locations = [0, 1.0]
			gradientLayer.startPoint = CGPoint(x: 0.0, y: 0.0)
			gradientLayer.endPoint = CGPoint(x: 1.0, y: 1.0)
			gradientLayer.borderWidth = 0
	
		case .solidDarkGreen:
			let lightColors = [
				UIColor(red: 42 / 255, green: 142 / 255, blue: 75 / 255, alpha: 1).cgColor,
				UIColor(red: 42 / 255, green: 142 / 255, blue: 75 / 255, alpha: 1).cgColor
			]

			let darkColors = [
				UIColor(red: 31 / 255, green: 89 / 255, blue: 50 / 255, alpha: 1).cgColor,
				UIColor(red: 31 / 255, green: 89 / 255, blue: 50 / 255, alpha: 1).cgColor
			]

			gradientLayer.colors = isDarkMode ? lightColors : darkColors
			gradientLayer.locations = [0, 1.0]
			gradientLayer.startPoint = CGPoint(x: 0.0, y: 0.0)
			gradientLayer.endPoint = CGPoint(x: 1.0, y: 1.0)
			gradientLayer.borderWidth = 0
			
		case .green:
			gradientLayer.colors = [
				UIColor(red: 46 / 255, green: 133 / 255, blue: 75 / 255, alpha: 1).cgColor,
				UIColor(red: 41 / 255, green: 117 / 255, blue: 67 / 255, alpha: 1).cgColor
			]
			gradientLayer.locations = [0.0, 0.95, 1.0]
			gradientLayer.startPoint = CGPoint(x: 0.25, y: 0.5)
			gradientLayer.endPoint = CGPoint(x: 0.75, y: 0.5)
			
		case .whiteWithGreyBorder:
			let lightColors = [
				UIColor(red: 255 / 255, green: 255 / 255, blue: 255 / 255, alpha: 1).cgColor,
				UIColor(red: 255 / 255, green: 255 / 255, blue: 255 / 255, alpha: 1).cgColor
			]

			let darkColors = [
				UIColor(red: 67 / 255, green: 68 / 255, blue: 69 / 255, alpha: 1).cgColor,
				UIColor(red: 67 / 255, green: 68 / 255, blue: 69 / 255, alpha: 1).cgColor
			]
			
			gradientLayer.colors = isDarkMode ? darkColors : lightColors
			gradientLayer.locations = [0, 1.0]
			gradientLayer.startPoint = CGPoint(x: 0.0, y: 0.0)
			gradientLayer.endPoint = CGPoint(x: 1.0, y: 1.0)
			
			gradientLayer.borderColor = isDarkMode ? UIColor(red: 255 / 255, green: 255 / 255, blue: 255 / 255, alpha: 1).cgColor : UIColor(red: 80 / 255, green: 89 / 255, blue: 104 / 255, alpha: 1).cgColor
			gradientLayer.borderWidth = 2
			
		case .whiteToLightBlue:
			let lightColors = [
				UIColor(red: 255 / 255, green: 255 / 255, blue: 255 / 255, alpha: 1).cgColor,
				UIColor(red: 255 / 255, green: 255 / 255, blue: 255 / 255, alpha: 1).cgColor,
				UIColor(red: 235 / 255, green: 244 / 255, blue: 255 / 255, alpha: 1).cgColor
			]

			let darkColors = [
				UIColor(red: 25 / 255, green: 25 / 255, blue: 27 / 255, alpha: 1).cgColor,
				UIColor(red: 47 / 255, green: 65 / 255, blue: 77 / 255, alpha: 1).cgColor
			]

			gradientLayer.colors = isDarkMode ? darkColors : lightColors
			gradientLayer.locations = [0.0, 0.6, 1.0]
			gradientLayer.startPoint = CGPoint(x: 0.0, y: 0.0)
			gradientLayer.endPoint = CGPoint(x: 1.0, y: 0.95)
		
		case .lightBlueToWhite:
			let lightColors = [
				UIColor(red: 255 / 255, green: 255 / 255, blue: 255 / 255, alpha: 1).cgColor,
				UIColor(red: 255 / 255, green: 255 / 255, blue: 255 / 255, alpha: 1).cgColor,
				UIColor(red: 227 / 255, green: 243 / 255, blue: 255 / 255, alpha: 1).cgColor
			]
			
			let darkColors = [
				UIColor(red: 30 / 255, green: 30 / 255, blue: 31 / 255, alpha: 1).cgColor,
				UIColor(red: 45 / 255, green: 65 / 255, blue: 77 / 255, alpha: 1).cgColor
			]
			
			gradientLayer.colors = isDarkMode ? darkColors : lightColors
			gradientLayer.locations = isDarkMode ? [0, 1] : [0, 0.34, 0.85]
			gradientLayer.startPoint = CGPoint(x: 0.25, y: 0)
			gradientLayer.endPoint = CGPoint(x: 0.75, y: 1)
		}
	}

	private func updateLayer() {
		updateStars()
		updateGradient()
	}

}
